(* Some of the errors that Links produces, and ways of displaying them. *)

type synerrspec = {filename : string; linespec : string; 
                   message : string; linetext : string;
                   marker : string}

exception NoMainExpr
exception ManyMainExprs of Syntax.expression list
exception MultiplyDefinedToplevelNames of 
  ((Expression.position list) Utility.stringmap)
exception Type_error of (Expression.position * string)
exception RichSyntaxError of synerrspec


val mistyped_application : Expression.position ->
  (Syntax.expression * Expression.Types.datatype) ->
  (Syntax.expression list * Expression.Types.datatype list) ->
  Expression.Types.datatype option ->
  'a
               
val mistyped_union : Expression.position ->
  Syntax.expression ->
  Expression.Types.datatype ->
  Syntax.expression ->
  Expression.Types.datatype ->
  'a

val mistype : Expression.position ->
  Syntax.expression * Expression.Types.datatype ->
  Expression.Types.datatype ->
  'a              

val nested_def : Expression.position -> string -> 'a
  
  
val letrec_nonfunction : Expression.position -> 
  (Syntax.expression * Expression.Types.datatype) ->
  'a

val invalid_name : Expression.position -> string -> string -> 'a

val format_exception : exn -> string
val format_exception_html : exn -> string

val display_fatal : ?stream:out_channel -> ('a -> 'b) -> ('a -> 'b)
val display_fatal_l : ?stream:out_channel -> ('b lazy_t) -> 'b
val display : ?default:(exn -> 'a) -> ?stream:out_channel -> ('a lazy_t) -> 'a
